import torch
from .heaviside import heaviside


class SuperSpike(torch.autograd.Function):
    """SuperSpike surrogate gradient as described in Section 3.3.2 of
    
        F. Zenke, S. Ganguli, "SuperSpike: Supervised Learning in Multilayer Spiking Neural Networks", 
        Neural Computation 30, 1514–1541 (2018),
        doi:10.1162/neco_a_01086
    """

    @staticmethod
    def forward(ctx, input: torch.Tensor, alpha: float) -> torch.Tensor:
        ctx.save_for_backward(input)
        ctx.alpha = alpha
        return heaviside(input)

    @staticmethod
    def backward(ctx, grad_output):
        (input,) = ctx.saved_tensors
        alpha = ctx.alpha
        grad_input = grad_output.clone()
        grad = grad_input / (alpha * torch.abs(input) + 1.0).pow(
            2
        )  # section 3.3.2 (beta -> alpha)
        return grad, None


def super_fn(x: torch.Tensor, alpha: float = 100.0) -> torch.Tensor:
    return SuperSpike.apply(x, alpha)



import torch

from ..functional.if_current_encoder import if_current_encoder_step


class IFConstantCurrentEncoder(torch.nn.Module):
    def __init__(
        self,
        seq_length,
        tau_mem_inv=1.0 / 1e-2,
        v_th=1.0,
        v_reset=0.0,
        dt: float = 0.001,
        device="cpu",
    ):
        super(IFConstantCurrentEncoder, self).__init__()
        self.seq_length = seq_length
        self.tau_mem_inv = tau_mem_inv
        self.v_th = v_th
        self.v_reset = v_reset
        self.device = device

    def forward(self, x):
        v = torch.zeros(*x.shape, device=self.device)
        z = torch.zeros(*x.shape, device=self.device)
        voltages = torch.zeros(self.seq_length, *x.shape, device=self.device)
        spikes = torch.zeros(self.seq_length, *x.shape, device=self.device)

        for ts in range(self.seq_length):
            z, v = if_current_encoder_step(
                input_current=x,
                v=v,
                tau_mem_inv=self.tau_mem_inv,
                v_th=self.v_th,
                v_reset=self.v_reset,
            )
            voltages[ts, :, :] = v
            spikes[ts, :, :] = z
        return spikes, voltages



from typing import Tuple
import numpy as np
import torch

from ..functional.lif import LIFState, LIFFeedForwardState, LIFParameters

from ..functional.lif_refrac import (
    LIFRefracParameters,
    LIFRefracState,
    LIFRefracFeedForwardState,
    lif_refrac_step,
    lif_refrac_feed_forward_step,
)


class LIFRefracCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a LIF neuron-model
    with absolute refractory period. More specifically it implements one integration 
    step of the following ODE.

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\\tau_{\\text{mem}} (1-\Theta(\\rho)) (v_{\\text{leak}} - v + i) \\\\
            \dot{i} &= -1/\\tau_{\\text{syn}} i \\\\
            \dot{\\rho} &= -1/\\tau_{\\text{refrac}} \Theta(\\rho)
        \end{align*}

    together with the jump condition

    .. math::
        \\begin{align*}
            z &= \Theta(v - v_{\\text{th}}) \\\\
            z_r &= \Theta(-\\rho)
        \end{align*}

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\\\
            i &= i + w_{\\text{rec}} z_{\\text{rec}} \\\\
            \\rho &= \\rho + z_r \\rho_{\\text{reset}} 
        \end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the recurrent and input
    spikes respectively.

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFRefracState): state at the current time step
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use

    Examples:

        >>> batch_size = 16
        >>> lif = LIFRefracCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> s0 = lif.initial_state(batch_size)
        >>> output, s0 = lif(input, s0)
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        p: LIFRefracParameters = LIFRefracParameters(),
        dt: float = 0.001,
    ):
        super(LIFRefracCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def initial_state(self, batch_size, device, dtype=torch.float) -> LIFRefracState:
        return LIFRefracState(
            lif=LIFState(
                z=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
                v=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
                i=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            ),
            rho=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LIFRefracState
    ) -> Tuple[torch.Tensor, LIFRefracState]:
        return lif_refrac_step(
            input,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LIFRefracFeedForwardCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a LIF neuron-model
    with absolute refractory period. More specifically it implements one integration 
    step of the following ODE.

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\\tau_{\\text{mem}} (1-\Theta(\\rho)) (v_{\\text{leak}} - v + i) \\\\
            \dot{i} &= -1/\\tau_{\\text{syn}} i \\\\
            \dot{\\rho} &= -1/\\tau_{\\text{refrac}} \Theta(\\rho)
        \end{align*}

    together with the jump condition

    .. math::
        \\begin{align*}
            z &= \Theta(v - v_{\\text{th}}) \\\\
            z_r &= \Theta(-\\rho)
        \end{align*}

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            \\rho &= \\rho + z_r \\rho_{\\text{reset}} 
        \end{align*}

    Parameters:
        shape: Shape of the processed spike input
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use

    Examples:
        >>> batch_size = 16
        >>> lif = LIFRefracFeedForwardCell((20, 30))
        >>> input = torch.randn(batch_size, 20, 30)
        >>> s0 = lif.initial_state(batch_size)
        >>> output, s0 = lif(input, s0)
    """

    def __init__(
        self, shape, p: LIFRefracParameters = LIFRefracParameters(), dt: float = 0.001
    ):
        super(LIFRefracFeedForwardCell, self).__init__()
        self.shape = shape
        self.p = p
        self.dt = dt

    def initial_state(self, batch_size, device, dtype) -> LIFFeedForwardState:
        return LIFRefracFeedForwardState(
            LIFFeedForwardState(
                v=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
                i=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
            ),
            rho=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LIFRefracFeedForwardState
    ) -> Tuple[torch.Tensor, LIFRefracFeedForwardState]:
        return lif_refrac_feed_forward_step(input, state, p=self.p, dt=self.dt)



from ..functional.lif import (
    LIFState,
    LIFFeedForwardState,
    LIFParameters,
    lif_step,
    lif_feed_forward_step,
    lif_current_encoder,
)

import torch
import time
import numpy as np


def lif_benchmark(
        input_features=10,
        output_features=10,
        n_time_steps=100,
        batch_size=16,
        input_spikes=np.zeros((100, 16, 10))
):
    iw = torch.randn(output_features, input_features)
    rw = torch.randn(output_features, output_features)
    T = n_time_steps
    s = LIFState(
        z=torch.zeros(batch_size, output_features),
        v=torch.zeros(batch_size, output_features),
        i=torch.zeros(batch_size, output_features)
    )
    p = LIFParameters(
        alpha=100.0,
        method="heaviside"
    )

    start = time.time()
    for ts in range(T):
        z, s = lif_step(
            input=input_spikes[ts, :],
            s=s,
            input_weights=iw,
            recurrent_weights=rw,
            p=p,
            dt=0.001
        )

    end = time.time()
    dt = (end - start) / T
    return dt


def lif_feed_forward_benchmark(
        input_features=10,
        output_features=10,
        n_time_steps=100,
        batch_size=16,
        input_spikes=np.zeros((100, 16, 10))
):
    fc = torch.nn.Linear(input_features, output_features, bias=False)
    T = n_time_steps
    s = LIFFeedForwardState(
        v=torch.zeros(batch_size, output_features),
        i=torch.zeros(batch_size, output_features)
    )
    p = LIFParameters(
        alpha=100.0,
        method="heaviside"
    )

    start = time.time()
    for ts in range(T):
        x = fc(input_spikes[ts, :])
        z, s = lif_feed_forward_step(
            input=x,
            s=s,
            p=p,
            dt=0.01
        )

    end = time.time()
    dt = (end - start) / T
    return dt



import torch


def heaviside(input):
    return torch.where(
        input <= torch.zeros_like(input),
        torch.zeros_like(input),
        torch.ones_like(input),
    )
    # return 0.5 * (torch.sign(input) + 1.0)



import torch
import numpy as np

from .lif import LIFFeedForwardState, LIFState
from .lif_refrac import LIFRefracState, LIFRefracFeedForwardState
from .lif_mc_refrac import lif_mc_refrac_step, lif_mc_refrac_feed_forward_step


def lif_refrac_step_test():
    input = torch.ones(20)
    s = LIFRefracState(
        lif=LIFState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10)),
        rho=torch.zeros(10),
    )
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()
    g_coupling = torch.randn(10, 10).float()

    for i in range(100):
        z, s = lif_mc_refrac_step(
            input, s, input_weights, recurrent_weights, g_coupling
        )


def lif_refrac_feed_forward_step_test():
    input = torch.ones(10)
    s = LIFRefracFeedForwardState(
        lif=LIFFeedForwardState(v=torch.zeros(10), i=torch.zeros(10)),
        rho=torch.zeros(10),
    )
    g_coupling = torch.randn(10, 10).float()

    for i in range(100):
        z, s = lif_mc_refrac_feed_forward_step(input, s, g_coupling)



import torch
import numpy as np

from .lif import (
    LIFState,
    LIFFeedForwardState,
    lif_step,
    lif_feed_forward_step,
    lif_current_encoder,
)


def lif_step_test():
    input = torch.ones(20)
    s = LIFState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10))
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for i in range(100):
        z, s = lif_step(input, s, input_weights, recurrent_weights)


def lif_feed_forward_step_test():
    input = torch.ones(10)
    s = LIFFeedForwardState(v=torch.zeros(10), i=torch.zeros(10))

    for i in range(100):
        z, s = lif_feed_forward_step(input, s)


def lif_current_encoder_test():
    input = torch.ones(10)
    v = torch.zeros(10)

    for i in range(100):
        z, v = lif_current_encoder(input, v)



from . import coba_lif
from . import if_current_encoder
from . import leaky_integrator
from . import lif_correlation
from . import lif_mc_refrac
from . import lif_mc
from . import lif
from . import lsnn



import torch

from .lif import LIFParameters, LIFState, LIFFeedForwardState
from .threshhold import threshhold

from typing import NamedTuple, Tuple


class LIFRefracState(NamedTuple):
    """State of a LIF neuron with absolute refractory period.

    Parameters:
        lif (LIFState): state of the LIF neuron integration
        rho (torch.Tensor): refractory state (count towards zero)
    """

    lif: LIFState
    rho: torch.Tensor


class LIFRefracParameters(NamedTuple):
    """Parameters of a LIF neuron with absolute refractory period.

    Parameters:
        lif (LIFParameters): parameters of the LIF neuron integration
        rho (torch.Tensor): refractory state (count towards zero)
    """

    lif: LIFParameters = LIFParameters()
    rho_reset: torch.Tensor = torch.as_tensor(5.0)


def lif_refrac_step(
    input: torch.Tensor,
    s: LIFRefracState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LIFRefracParameters = LIFRefracParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFRefracState]:
    r"""Computes a single euler-integration step of a recurrently connected
     LIF neuron-model with a refractory period.
    
    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFRefracState): state at the current time step
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use
    """
    refrac_mask = threshhold(s.rho, p.lif.method, p.lif.alpha)

    # compute voltage updates
    dv = (
        (1 - refrac_mask)
        * dt
        * p.lif.tau_mem_inv
        * ((p.lif.v_leak - s.lif.v) + s.lif.i)
    )
    v_decayed = s.lif.v + dv

    # compute current updates
    di = -dt * p.lif.tau_syn_inv * s.lif.i
    i_decayed = s.lif.i + di

    # compute new spikes
    z_new = threshhold(v_decayed - p.lif.v_th, p.lif.method, p.lif.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.lif.v_reset
    # compute current jumps
    i_new = (
        i_decayed
        + torch.nn.functional.linear(input, input_weights)
        + torch.nn.functional.linear(s.lif.z, recurrent_weights)
    )

    # compute update to refractory counter
    rho_new = (1 - z_new) * torch.nn.functional.relu(
        s.rho - refrac_mask
    ) + z_new * p.rho_reset

    return z_new, LIFRefracState(LIFState(z_new, v_new, i_new), rho_new)


class LIFRefracFeedForwardState(NamedTuple):
    """State of a feed forward LIF neuron with absolute refractory period.

    Parameters:
        lif (LIFFeedForwardState): state of the feed forward LIF neuron integration
        rho (torch.Tensor): refractory state (count towards zero)
    """

    lif: LIFFeedForwardState
    rho: torch.Tensor


def lif_refrac_feed_forward_step(
    input: torch.Tensor,
    s: LIFRefracFeedForwardState,
    p: LIFRefracParameters = LIFRefracParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFRefracFeedForwardState]:
    r"""Computes a single euler-integration step of a feed forward
     LIF neuron-model with a refractory period.

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFRefracFeedForwardState): state at the current time step
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use
    """
    rho_mask = threshhold(s.rho, p.lif.method, p.lif.alpha)

    # compute voltage updates
    dv = (1 - rho_mask) * dt * p.lif.tau_mem_inv * ((p.lif.v_leak - s.lif.v) + s.lif.i)
    v_decayed = s.lif.v + dv

    # compute current updates
    di = -dt * p.lif.tau_syn_inv * s.lif.i
    i_decayed = s.lif.i + di

    # compute new spikes
    z_new = threshhold(v_decayed - p.lif.v_th, p.lif.method, p.lif.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.lif.v_reset
    # compute current jumps
    i_new = i_decayed + input

    # compute update to refractory counter
    rho_new = (1 - z_new) * torch.nn.functional.relu(
        s.rho - rho_mask
    ) + z_new * p.rho_reset
    return z_new, LIFRefracFeedForwardState(LIFFeedForwardState(v_new, i_new), rho_new)



from typing import Tuple

import torch
import numpy as np

from ..functional.lsnn import (
    LSNNParameters,
    LSNNState,
    LSNNFeedForwardState,
    lsnn_step,
    lsnn_feed_forward_step,
)


class LSNNCell(torch.nn.Module):
    r"""Module that computes a single euler-integration step of a LSNN neuron-model.
    More specifically it implements one integration step of the following ODE

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - v + i) \\\\
            \dot{i} &= -1/\\tau_{\\text{syn}} i \\\\
            \dot{b} &= -1/\\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\\text{th}} + b)

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\\\
            i &= i + w_{\\text{rec}} z_{\\text{rec}} \\\\
            b &= b + \\beta z
        \end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the recurrent and input
    spikes respectively.

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LSNNState): current state of the lsnn unit
        input_weights (torch.Tensor): synaptic weights for input spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """

    def __init__(
        self,
        input_features,
        output_features,
        p: LSNNParameters = LSNNParameters(),
        dt: float = 0.001,
    ):
        super(LSNNCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(output_features, input_features) / np.sqrt(input_features)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(output_features, output_features)
        )
        self.input_features = input_features
        self.output_features = output_features
        self.p = p
        self.dt = dt

    def initial_state(self, batch_size, device, dtype=torch.float) -> LSNNState:
        """return the initial state of an LSNN neuron"""
        return LSNNState(
            z=torch.zeros(batch_size, self.output_features, device=device, dtype=dtype),
            v=torch.zeros(batch_size, self.output_features, device=device, dtype=dtype),
            i=torch.zeros(batch_size, self.output_features, device=device, dtype=dtype),
            b=torch.zeros(batch_size, self.output_features, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LSNNState
    ) -> Tuple[torch.Tensor, LSNNState]:
        return lsnn_step(
            input,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LSNNLayer(torch.nn.Module):
    r"""A Long short-term memory neuron module adapted from https://arxiv.org/abs/1803.09574

    Usage:
      >>> from norse.torch.module import LSNNLayer, LSNNCell
      >>> layer = LSNNLayer(LSNNCell, 2, 10)    // LSNNCell with 2 inputs and 10 outputs
      >>> state = layer.initial_state(5, "cpu") // 5 batch size running on CPU
      >>> data  = torch.zeros(2, 5, 2)          // Generate data of shape [5, 2, 10]
      >>> output, new_state = layer.forward(data, state)

    Parameters:
      cell (torch.nn.Module): the underling neuron module, uninitialized
      *cell_args: variable length input arguments for the underlying cell constructor
    """

    def __init__(self, cell, *cell_args):
        super(LSNNLayer, self).__init__()
        self.cell = cell(*cell_args)

    def initial_state(self, batch_size, device, dtype=torch.float) -> LSNNState:
        """Return the initial state of the LSNN layer, as given by the internal LSNNCell"""
        return self.cell.initial_state(batch_size, device, dtype)

    def forward(
        self, input: torch.Tensor, state: LSNNState
    ) -> Tuple[torch.Tensor, LSNNState]:
        inputs = input.unbind(0)
        outputs = []  # torch.jit.annotate(List[torch.Tensor], [])
        for i in range(len(inputs)):
            out, state = self.cell(inputs[i], state)
            outputs += [out]
        return torch.stack(outputs), state


class LSNNFeedForwardCell(torch.nn.Module):
    r"""Euler integration cell for LIF Neuron with threshhold adaptation.
    More specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\\text{syn}} i \\
            \dot{b} &= -1/\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}} + b)

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + \text{input} \\
            b &= b + \beta z
        \end{align*}

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LSNNFeedForwardState): current state of the lsnn unit
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """

    def __init__(self, shape, p: LSNNParameters = LSNNParameters(), dt: float = 0.001):
        super(LSNNFeedForwardCell, self).__init__()
        self.shape = shape
        self.p = p
        self.dt = dt

    def initial_state(
        self, batch_size, device, dtype=torch.float
    ) -> LSNNFeedForwardState:
        """return the initial state of an LSNN neuron"""
        return LSNNFeedForwardState(
            v=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
            i=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
            b=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LSNNFeedForwardState
    ) -> Tuple[torch.Tensor, LSNNFeedForwardState]:
        return lsnn_feed_forward_step(input, state, p=self.p, dt=self.dt)



import torch

from ..functional.lif import LIFState, LIFParameters
from ..functional.lif_mc import lif_mc_step

import numpy as np
from typing import Tuple


class LIFMCCell(torch.nn.Module):
    """Computes a single euler-integration step of a LIF multi-compartment
    neuron-model.

    .. math::
        \\begin{align*}
            \\dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - g_{\\text{coupling}} v + i) \\\\
            \\dot{i} &= -1/\\tau_{\\text{syn}} i
        \end{align*}

    together with the jump condition
    
    .. math::
        z = \Theta(v - v_{\\text{th}})
    
    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\\\
            i &= i + w_{\\text{rec}} z_{\\text{rec}}
        \end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the recurrent and input
    spikes respectively.
    

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFState): current state of the neuron
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        g_coupling (torch.Tensor): conductances between the neuron compartments
        p (LIFParameters): neuron parameters
        dt (float): Integration timestep to use
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFParameters = LIFParameters(),
        dt: float = 0.001,
    ):
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.g_coupling = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.p = p
        self.dt = dt

    def initial_state(
        self, batch_size: int, device: torch.device, dtype=torch.float
    ) -> LIFState:
        return LIFState(
            z=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            v=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            i=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LIFState
    ) -> Tuple[torch.Tensor, LIFState]:
        return lif_mc_step(
            input,
            state,
            self.input_weights,
            self.recurrent_weights,
            self.g_coupling,
            p=self.p,
            dt=self.dt,
        )



import numpy as np

def benchmark(
        run,
        n_runs=100,
        **kwargs
):
    dts = []
    for i in range(n_runs):
        dt = run(**kwargs)
        dts.append(dt)

    return np.stack(dts)



import torch

from ..functional.coba_lif import CobaLIFParameters, CobaLIFState, coba_lif_step

from typing import Tuple
import numpy as np


class CobaLIFCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a conductance based
    LIF neuron-model. More specifically it implements one integration step of the following ODE

    .. math::
        \\begin{align*}
            \dot{v} &= 1/c_{\\text{mem}} (g_l (v_{\\text{leak}} - v) + g_e (E_{\\text{rev_e}} - v) + g_i (E_{\\text{rev_i}} - v)) \\\\
            \dot{g_e} &= -1/\\tau_{\\text{syn}} g_e \\\\
            \dot{g_i} &= -1/\\tau_{\\text{syn}} g_i
        \end{align*}

    together with the jump condition
    
    .. math::
        z = \Theta(v - v_{\\text{th}})
    
    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            g_e &= g_e + \\text{relu}(w_{\\text{input}}) z_{\\text{in}} \\\\
            g_e &= g_e + \\text{relu}(w_{\\text{rec}}) z_{\\text{rec}} \\\\
            g_i &= g_i + \\text{relu}(-w_{\\text{input}}) z_{\\text{in}} \\\\
            g_i &= g_i + \\text{relu}(-w_{\\text{rec}}) z_{\\text{rec}} \\\\
        \end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the recurrent and input
    spikes respectively.

    Parameters:
        input_size (int): Size of the input.
        hidden_size (int): Size of the hidden state.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif = CobaLIFCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> s0 = lif.initial_state(batch_size)
        >>> output, s0 = lif(input, s0)
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        p: CobaLIFParameters = CobaLIFParameters(),
        dt: float = 0.001,
    ):
        super(CobaLIFCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.p = p
        self.dt = dt

    def initial_state(
        self, batch_size: int, device: torch.device, dtype=torch.float
    ) -> CobaLIFState:
        return CobaLIFState(
            z=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            v=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            g_e=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            g_i=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: CobaLIFState
    ) -> Tuple[torch.Tensor, CobaLIFState]:
        return coba_lif_step(
            input,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )



import torch

from .heaviside import heaviside


def if_current_encoder_step(
    input_current, v, tau_mem_inv=1.0 / 1e-2, v_th=1.0, v_reset=0.0, dt=0.001
):
    dv = dt * tau_mem_inv * (input_current)
    v = v + dv
    z = heaviside(v - v_th)

    v = v - z * (v - v_reset)
    return z, v



import torch

from ..functional.lif_correlation import (
    LIFCorrelationState,
    LIFCorrelationParameters,
    lif_correlation_step,
)
from ..functional.lif import LIFState
from ..functional.correlation_sensor import CorrelationSensorState

from typing import Tuple


class LIFCorrelation(torch.nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        p: LIFCorrelationParameters = LIFCorrelationParameters(),
        dt: float = 0.001,
    ):
        super(LIFCorrelation, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.p = p
        self.dt = dt

    def initial_state(
        self, batch_size, device, dtype=torch.float
    ) -> LIFCorrelationState:
        hidden_features = self.hidden_size
        input_features = self.input_size

        return LIFCorrelationState(
            lif_state=LIFState(
                z=torch.zeros(batch_size, hidden_features, device=device, dtype=dtype),
                v=torch.zeros(batch_size, hidden_features, device=device, dtype=dtype),
                i=torch.zeros(batch_size, hidden_features, device=device, dtype=dtype),
            ),
            input_correlation_state=CorrelationSensorState(
                post_pre=torch.zeros(
                    (batch_size, input_features, hidden_features),
                    device=device,
                    dtype=dtype,
                ),
                correlation_trace=torch.zeros(
                    (batch_size, input_features, hidden_features),
                    device=device,
                    dtype=dtype,
                ).float(),
                anti_correlation_trace=torch.zeros(
                    (batch_size, input_features, hidden_features),
                    device=device,
                    dtype=dtype,
                ).float(),
            ),
            recurrent_correlation_state=CorrelationSensorState(
                correlation_trace=torch.zeros(
                    (batch_size, hidden_features, hidden_features),
                    device=device,
                    dtype=dtype,
                ),
                anti_correlation_trace=torch.zeros(
                    (batch_size, hidden_features, hidden_features),
                    device=device,
                    dtype=dtype,
                ),
                post_pre=torch.zeros(
                    (batch_size, hidden_features, hidden_features),
                    device=device,
                    dtype=dtype,
                ),
            ),
        )

    def forward(
        self,
        input: torch.Tensor,
        s: LIFCorrelationState,
        input_weights: torch.Tensor,
        recurrent_weights: torch.Tensor,
    ) -> Tuple[torch.Tensor, LIFCorrelationState]:
        return lif_correlation_step(
            input, s, input_weights, recurrent_weights, self.p, self.dt
        )



import torch

from ..functional.lif_refrac import LIFRefracParameters, LIFRefracState
from ..functional.lif import LIFState
from ..functional.lif_mc_refrac import lif_mc_refrac_step

import numpy as np
from typing import Tuple


class LIFMCRefracCell(torch.nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFRefracParameters = LIFRefracParameters(),
        dt: float = 0.001,
    ):
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.g_coupling = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.p = p
        self.dt = dt

    def initial_state(
        self, batch_size: int, device: torch.device, dtype: torch.float = torch.float
    ) -> LIFRefracState:
        return LIFRefracState(
            lif=LIFState(
                z=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
                v=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
                i=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            ),
            rho=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LIFRefracState
    ) -> Tuple[torch.Tensor, LIFRefracState]:
        return lif_mc_refrac_step(
            input,
            state,
            self.input_weights,
            self.recurrent_weights,
            self.g_coupling,
            p=self.p,
            dt=self.dt,
        )



from typing import Tuple, List

import numpy as np
import torch

from ..functional.lif import (
    LIFState,
    LIFFeedForwardState,
    LIFParameters,
    lif_step,
    lif_feed_forward_step,
    lif_current_encoder,
)


class LIFCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a LIF neuron-model. 
    More specifically it implements one integration step of the following ODE

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - v + i) \\\\
            \dot{i} &= -1/\\tau_{\\text{syn}} i
        \end{align*}

    together with the jump condition
    
    .. math::
        z = \Theta(v - v_{\\text{th}})
    
    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\\\
            i &= i + w_{\\text{rec}} z_{\\text{rec}}
        \end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the recurrent and input
    spikes respectively.

    Parameters:
        input_size (int): Size of the input.
        hidden_size (int): Size of the hidden state.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif = LIFCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> s0 = lif.initial_state(batch_size)
        >>> output, s0 = lif(input, s0)
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        p: LIFParameters = LIFParameters(),
        dt: float = 0.001,
    ):
        super(LIFCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) * np.sqrt(2 / hidden_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        )
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def extra_repr(self):
        s = f"{self.input_size}, {self.hidden_size}, p={self.p}, dt={self.dt}"
        return s

    def initial_state(self, batch_size, device, dtype=torch.float) -> LIFState:
        return LIFState(
            z=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            v=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
            i=torch.zeros(batch_size, self.hidden_size, device=device, dtype=dtype),
        )

    def forward(
        self, input: torch.Tensor, state: LIFState
    ) -> Tuple[torch.Tensor, LIFState]:
        return lif_step(
            input,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LIFLayer(torch.nn.Module):
    def __init__(self, *cell_args, **kw_args):
        super(LIFLayer, self).__init__()
        self.cell = LIFCell(*cell_args, **kw_args)

    def forward(
        self, input: torch.Tensor, state: LIFState
    ) -> Tuple[torch.Tensor, LIFState]:
        inputs = input.unbind(0)
        outputs = []  # torch.jit.annotate(List[torch.Tensor], [])
        for i in range(len(inputs)):
            out, state = self.cell(inputs[i], state)
            outputs += [out]
        return torch.stack(outputs), state


class LIFFeedForwardCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a LIF neuron.
    It takes as input the input current as generated by an arbitrary torch module
    or function. More specifically it implements one integration step of the following ODE

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - v + i) \\\\
            \dot{i} &= -1/\\tau_{\\text{syn}} i
        \end{align*}

    together with the jump condition
    
    .. math::
        z = \Theta(v - v_{\\text{th}})
    
    and transition equations

    .. math::
        i = i + i_{\\text{in}}

    where :math:`i_{\\text{in}}` is meant to be the result of applying an arbitrary
    pytorch module (such as a convolution) to input spikes.

    Parameters:
        shape: Shape of the feedforward state.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif = LIFFeedForwardCell((20, 30))
        >>> data = torch.randn(batch_size, 20, 30)
        >>> s0 = lif.initial_state(batch_size, "cpu")
        >>> output, s0 = lif(data, s0)
    """

    def __init__(self, shape, p: LIFParameters = LIFParameters(), dt: float = 0.001):
        super(LIFFeedForwardCell, self).__init__()
        self.shape = shape
        self.p = p
        self.dt = dt

    def extra_repr(self):
        s = f"{self.shape}, p={self.p}, dt={self.dt}"
        return s

    def initial_state(self, batch_size, device, dtype=None) -> LIFFeedForwardState:
        return LIFFeedForwardState(
            v=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
            i=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
        )

    def forward(
        self, x: torch.Tensor, state: LIFFeedForwardState
    ) -> Tuple[torch.Tensor, LIFFeedForwardState]:
        return lif_feed_forward_step(x, state, p=self.p, dt=self.dt)


class LIFConstantCurrentEncoder(torch.nn.Module):
    def __init__(
        self,
        seq_length,
        p: LIFParameters = LIFParameters(),
        dt: float = 0.001,
        device="cpu",
    ):
        super(LIFConstantCurrentEncoder, self).__init__()
        self.seq_length = seq_length
        self.p = p
        self.device = device
        self.dt = dt

    def forward(self, x):
        v = torch.zeros(*x.shape, device=self.device)
        z = torch.zeros(*x.shape, device=self.device)
        voltages = torch.zeros(self.seq_length, *x.shape, device=self.device)
        spikes = torch.zeros(self.seq_length, *x.shape, device=self.device)

        for ts in range(self.seq_length):
            z, v = lif_current_encoder(input_current=x, v=v, p=self.p, dt=self.dt)
            voltages[ts, :, :] = v
            spikes[ts, :, :] = z
        return voltages, spikes



import torch
import torch.jit
import numpy as np

from ..functional.leaky_integrator import (
    li_step,
    li_feed_forward_step,
    LIState,
    LIParameters,
)

from typing import Tuple


class LICell(torch.nn.Module):
    r"""Cell for a leaky-integrator.
    More specifically it implements a discretized version of the ODE

    .. math::

        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}


    and transition equations

    .. math::
        i = i + w i_{\text{in}}

    Parameters:
        input_features (int); Input feature dimension
        output_features (int): Output feature dimension
        p (LIParameters): parameters of the leaky integrator
        dt (float): integration timestep to use
    """

    def __init__(
        self,
        input_features: int,
        output_features: int,
        p: LIParameters = LIParameters(),
        dt: float = 0.001,
    ):
        super(LICell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(output_features, input_features) / np.sqrt(input_features)
        )
        self.p = p
        self.dt = dt
        self.output_features = output_features

    def initial_state(self, batch_size, device, dtype=torch.float) -> LIState:
        return LIState(
            v=torch.zeros(self.output_features, device=device, dtype=dtype),
            i=torch.zeros(self.output_features, device=device, dtype=dtype),
        )

    def forward(self, input: torch.Tensor, s: LIState) -> Tuple[torch.Tensor, LIState]:
        return li_step(input, s, self.input_weights, p=self.p, dt=self.dt)


class LIFeedForwardCell(torch.nn.Module):
    r"""Cell for a leaky-integrator.
    More specifically it implements a discretized version of the ODE

    .. math::

        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}


    and transition equations

    .. math::
        i = i + w i_{\text{in}}

    Parameters:
        shape: Shape of the preprocessed input spikes
        p (LIParameters): parameters of the leaky integrator
        dt (float): integration timestep to use
    """

    def __init__(self, shape, p: LIParameters = LIParameters(), dt: float = 0.001):
        super(LIFeedForwardCell, self).__init__()
        self.p = p
        self.dt = dt
        self.shape = shape

    def initial_state(self, batch_size, device, dtype=torch.float):
        return LIState(
            v=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
            i=torch.zeros(batch_size, *self.shape, device=device, dtype=dtype),
        )

    def forward(self, input: torch.Tensor, s: LIState) -> Tuple[torch.Tensor, LIState]:
        return li_feed_forward_step(input, s, p=self.p, dt=self.dt)



import torch
from typing import Tuple

from .lif import LIFState, LIFFeedForwardState, LIFParameters
from .threshhold import threshhold


def lif_mc_step(
    input: torch.Tensor,
    s: LIFState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    g_coupling: torch.Tensor,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFState]:
    """Computes a single euler-integration step of a LIF multi-compartment
    neuron-model.

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFState): current state of the neuron
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        g_coupling (torch.Tensor): conductances between the neuron compartments
        p (LIFParameters): neuron parameters
        dt (float): Integration timestep to use
    """
    # compute voltage
    dv = dt * p.tau_mem_inv * ((p.v_leak - s.v) + s.i) + torch.nn.functional.linear(
        s.v, g_coupling
    )
    v_decayed = s.v + dv
    # compute current updates
    di = -dt * p.tau_syn_inv * s.i
    i_decayed = s.i + di
    # compute new spikes
    z_new = threshhold(v_decayed - p.v_th, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = (
        i_decayed
        + torch.nn.functional.linear(input, input_weights)
        + torch.nn.functional.linear(s.z, recurrent_weights)
    )
    return z_new, LIFState(z_new, v_new, i_new)


def lif_mc_feed_forward_step(
    input: torch.Tensor,
    s: LIFFeedForwardState,
    g_coupling: torch.Tensor,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFFeedForwardState]:
    """Computes a single euler-integration feed forward step of a LIF multi-compartment
    neuron-model.

    Parameters:
        input (torch.Tensor): the (weighted) input spikes at the current time step
        s (LIFFeedForwardState): current state of the neuron
        g_coupling (torch.Tensor): conductances between the neuron compartments
        p (LIFParameters): neuron parameters
        dt (float): Integration timestep to use
    """
    # compute voltage
    dv = dt * p.tau_mem_inv * ((p.v_leak - s.v) + s.i) + torch.nn.functional.linear(
        s.v, g_coupling
    )
    v_decayed = s.v + dv
    # compute current updates
    di = -dt * p.tau_syn_inv * s.i
    i_decayed = s.i + di
    # compute new spikes
    z_new = threshhold(v_decayed - p.v_th, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = i_decayed + input
    return z_new, LIFFeedForwardState(v_new, i_new)



import torch
import numpy as np

from .leaky_integrator import LIState, li_feed_forward_step, li_step


def lif_step_test():
    input = torch.ones(20)
    s = LIState(v=torch.zeros(10), i=torch.zeros(10))
    input_weights = torch.randn(10, 20).float()

    for i in range(100):
        z, s = li_step(input, s, input_weights)


def lif_feed_forward_step_test():
    input = torch.ones(10)
    s = LIState(v=torch.zeros(10), i=torch.zeros(10))

    for i in range(100):
        z, s = li_feed_forward_step(input, s)



import os
import datetime
import uuid

from absl import app
from absl import flags
from absl import logging

import torch

from collections import namedtuple
import torchvision
import numpy as np
import matplotlib.pyplot as plt

from norse.torch.models.conv import ConvNet, ConvNet4
from norse.torch.module.if_current_encoder import IFConstantCurrentEncoder

from torch.utils.tensorboard import SummaryWriter

FLAGS = flags.FLAGS

flags.DEFINE_bool(
    "only_first_spike", False, "Only one spike per input (latency coding)."
)
flags.DEFINE_bool("save_grads", False, "Save gradients of backward pass.")
flags.DEFINE_integer(
    "grad_save_interval", 10, "Interval for gradient saving of backward pass."
)
flags.DEFINE_string("prefix", "", "Prefix for save path to use.")
flags.DEFINE_enum(
    "encoding",
    "poisson",
    ["poisson", "constant", "constant_polar", "signed_poisson", "signed_constant"],
    "Encoding to use for input",
)

flags.DEFINE_enum(
    "net", "convnet4", ["convnet", "convnet4"], "Which network architecture to use"
)
flags.DEFINE_integer("plot_interval", 10, "Interval for plotting.")
flags.DEFINE_float("input_scale", 1, "Scaling factor for input current.")
flags.DEFINE_float("current_encoder_v_th", 1.0, "v_th for constant current encoder")
flags.DEFINE_bool("learning_rate_schedule", False, "Use a learning rate schedule")
flags.DEFINE_bool("find_learning_rate", False, "Find learning rate")


class PiecewiseLinear(namedtuple("PiecewiseLinear", ("batch_size", "knots", "vals"))):
    def step(self, optimizer, t):
        lr = np.interp([t], self.knots, self.vals)[0]
        for group in optimizer.param_groups:
            group["lr"] = lr / self.batch_size


def generate_poisson_trains(batch_size, num_trains, seq_length, freq):
    trains = np.random.rand(seq_length, batch_size, num_trains) < freq
    return torch.from_numpy(trains).float()


def add_luminance(images):
    return torch.cat(
        (
            images,
            torch.unsqueeze(
                0.2126 * images[0, :, :]
                + 0.7152 * images[1, :, :]
                + 0.0722 * images[2, :, :],
                0,
            ),
        ),
        0,
    )


def poisson_train(images, seq_length, device, rel_fmax=0.2):
    return (
        torch.rand(seq_length, *images.shape).float() < rel_fmax * images
    ).float()


def signed_poisson_train(images, seq_length, device, rel_fmax=0.2):
    return (
        torch.sign(images)
        * (
            torch.rand(seq_length, *images.shape).float().to(device)
            < rel_fmax * torch.abs(images)
        ).float()
    )


class LIFConvNet(torch.nn.Module):
    def __init__(self, num_channels, seq_length, model="super", device="cpu"):
        super(LIFConvNet, self).__init__()

        if FLAGS.net == "convnet":
            dtype = torch.float
            self.rsnn = ConvNet(
                device=device, num_channels=num_channels, feature_size=32, dtype=dtype
            )
        elif FLAGS.net == "convnet4":
            self.rsnn = ConvNet4(
                device=device, num_channels=num_channels, feature_size=32
            )
        self.device = device

    def forward(self, x):
        voltages = self.rsnn(x.to(self.device).permute(1, 0, 2, 3, 4))
        m, _ = torch.max(voltages, 0)
        log_p_y = torch.nn.functional.log_softmax(m, dim=1)
        return log_p_y


def train(
    model, device, train_loader, optimizer, epoch, lr_scheduler=None, writer=None
):
    model.train()
    losses = []
    train_batches = len(train_loader)
    step = train_batches * epoch

    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = torch.nn.functional.nll_loss(output, target)
        loss.backward()
        if FLAGS.save_grads and batch_idx % FLAGS.grad_save_interval == 0:
            for idx, p in enumerate(model.parameters()):
                np.save(f"param-{idx}-{epoch}-{batch_idx}-grad.npy", p.grad.numpy())
                np.save(f"param-{idx}-{epoch}-{batch_idx}-data.npy", p.data.numpy())

        if lr_scheduler:
            lr_scheduler.step(optimizer, t=(epoch + batch_idx / train_batches))
        optimizer.step()
        step += 1

        if batch_idx % FLAGS.log_interval == 0:
            logging.info(
                "Train Epoch: {}/{} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    FLAGS.epochs,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                )
            )

        if step % FLAGS.log_interval == 0 and writer:
            _, argmax = torch.max(output, 1)
            accuracy = (target == argmax.squeeze()).float().mean()

            writer.add_scalar("Loss/train", loss.item(), step)
            writer.add_scalar("Accuracy/train", accuracy.item(), step)

            for tag, value in model.named_parameters():
                tag = tag.replace(".", "/")
                writer.add_histogram(tag, value.data.cpu().numpy(), step)
                writer.add_histogram(tag + "/grad", value.grad.data.cpu().numpy(), step)

        if FLAGS.do_plot and batch_idx % FLAGS.plot_interval == 0:
            ts = np.arange(0, FLAGS.seq_length) * FLAGS.dt
            _, axs = plt.subplots(4, 4, figsize=(15, 10), sharex=True, sharey=True)
            axs = axs.reshape(-1)  # flatten
            for nrn in range(10):
                one_trace = voltages.detach().cpu().numpy()[:, 0, nrn]
                plt.sca(axs[nrn])
                plt.plot(ts, one_trace)
            plt.xlabel("Time [s]")
            plt.ylabel("Membrane Potential")
            plt.show()
        losses.append(loss.item())

    mean_loss = np.mean(losses)
    return losses, mean_loss


def test(model, device, test_loader, epoch, writer=None):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += torch.nn.functional.nll_loss(
                output, target, reduction="sum"
            ).item()  # sum up batch loss
            pred = output.argmax(
                dim=1, keepdim=True
            )  # get the index of the max log-probabilioty
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    accuracy = 100.0 * correct / len(test_loader.dataset)
    logging.info(
        f"\nTest set {FLAGS.model}: Average loss: {test_loss:.4f}, Accuracy: {correct}/{len(test_loader.dataset)} ({accuracy:.0f}%)\n"
    )

    if writer:
        writer.add_scalar("Loss/test", test_loss, epoch)
        writer.add_scalar("Accuracy/test", accuracy, epoch)

    return test_loss, accuracy


def save(path, model, optimizer):
    torch.save(
        {
            "model_state_dict": model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
        },
        path,
    )


def load(path, model, optimizer, device):
    checkpoint = torch.load(path)
    model.load_state_dict(checkpoint["model_state_dict"])
    optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
    model.train(device=device)
    return model, optimizer


def compute_min(loader):
    min = 0.0
    for batch_idx, (data, target) in enumerate(loader):
        print(torch.min(data))


def main(argv):
    writer = SummaryWriter()

    torch.manual_seed(FLAGS.random_seed)

    np.random.seed(FLAGS.random_seed)
    if hasattr(torch, "cuda_is_available"):
        if torch.cuda_is_available():
            torch.cuda.manual_seed(FLAGS.random_seed)
            torch.backends.cudnn.enabled = True
            torch.backends.cudnn.benchmark = True

    device = torch.device(FLAGS.device)

    constant_current_encoder = IFConstantCurrentEncoder(
        seq_length=FLAGS.seq_length, v_th=FLAGS.current_encoder_v_th
    )

    def polar_current_encoder(x):
        x_p, _ = constant_current_encoder(2 * torch.nn.functional.relu(x))
        x_m, _ = constant_current_encoder(2 * torch.nn.functional.relu(-x))
        return torch.cat((x_p, x_m), 1)

    def current_encoder(x):
        x, _ = constant_current_encoder(2 * x)
        return x

    def identity_encoder(x):
        return x

    def poisson_encoder(x):
        return poisson_train(x, FLAGS.seq_length, device)

    def signed_poisson_encoder(x):
        return signed_poisson_train(x, FLAGS.seq_length, device)

    def signed_current_encoder(x):
        z, _ = constant_current_encoder(torch.abs(x))
        return torch.sign(x) * z

    num_channels = 4

    if FLAGS.encoding == "poisson":
        encoder = poisson_encoder
    elif FLAGS.encoding == "constant":
        encoder = current_encoder
    elif FLAGS.encoding == "signed_poisson":
        encoder = signed_poisson_encoder
    elif FLAGS.encoding == "signed_constant":
        encoder = signed_current_encoder
    elif FLAGS.encoding == "constant_polar":
        encoder = polar_current_encoder
        num_channels = 2 * num_channels

    def compute_min_max(loader):

        minimum = 0.0
        maximum = 0.0
        for data, _ in loader:
            minimum = min(torch.min(data), minimum)
            maximum = max(torch.max(data), maximum)
        return minimum, maximum

    def train_min_max():
        transform_train = torchvision.transforms.Compose(
            [torchvision.transforms.ToTensor(), add_luminance]
        )
        kwargs = (
            {"num_workers": 4, "pin_memory": True} if FLAGS.device is "cuda" else {}
        )
        train_loader = torch.utils.data.DataLoader(
            torchvision.datasets.CIFAR10(
                root=".", train=True, download=True, transform=transform_train
            ),
            batch_size=FLAGS.batch_size,
            shuffle=True,
            **kwargs,
        )
        return compute_min_max(train_loader)

    luminance_transforms = [
        add_luminance,
        torchvision.transforms.Normalize(
            (0.4914, 0.4822, 0.4465, 0.4816), (0.2023, 0.1994, 0.2010, 0.20013)
        ),
    ]

    transform_train = torchvision.transforms.Compose(
        [
            torchvision.transforms.RandomCrop(32, padding=4),
            torchvision.transforms.RandomHorizontalFlip(),
            torchvision.transforms.ToTensor(),
        ]
        + luminance_transforms
        + [encoder]
    )

    transform_test = torchvision.transforms.Compose(
        [torchvision.transforms.ToTensor()] + luminance_transforms + [encoder]
    )

    kwargs = {"num_workers": 0, "pin_memory": True} if FLAGS.device == "cuda" else {}
    train_loader = torch.utils.data.DataLoader(
        torchvision.datasets.CIFAR10(
            root=".", train=True, download=True, transform=transform_train
        ),
        batch_size=FLAGS.batch_size,
        shuffle=True,
        **kwargs,
    )
    test_loader = torch.utils.data.DataLoader(
        torchvision.datasets.CIFAR10(root=".", train=False, transform=transform_test),
        batch_size=FLAGS.batch_size,
        **kwargs,
    )

    label = os.environ.get("SLURM_JOB_ID", str(uuid.uuid4()))
    if not FLAGS.prefix:
        rundir = f"runs/cifar10/{label}"
    else:
        rundir = f"runs/cifar10/{FLAGS.prefix}/{label}"

    os.makedirs(rundir, exist_ok=True)
    os.chdir(rundir)
    FLAGS.append_flags_into_file(f"flags.txt")

    model = LIFConvNet(
        num_channels=num_channels,
        seq_length=FLAGS.seq_length,
        model=FLAGS.model,
        device=device,
    ).to(device)

    print(model)

    if device == "cuda":
        model = torch.nn.DataParallel(model).to(device)

    batch_size = FLAGS.batch_size

    if FLAGS.optimizer == "sgd":
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=FLAGS.learning_rate,
            momentum=0.9,
            weight_decay=5e-4 * FLAGS.batch_size,
            nesterov=True,
        )
    elif FLAGS.optimizer == "adam":
        optimizer = torch.optim.Adam(model.parameters(), lr=FLAGS.learning_rate)
    elif FLAGS.optimizer == "rms":
        optimizer = torch.optim.RMSprop(model.parameters(), lr=FLAGS.learning_rate)

    if FLAGS.only_output:
        optimizer = torch.optim.Adam(model.out.parameters(), lr=FLAGS.learning_rate)

    if FLAGS.resume:
        if os.path.isfile(FLAGS.resume):
            checkpoint = torch.load(FLAGS.resume)
            model.load_state_dict(checkpoint["state_dict"])
            optimizer.load_state_dict(checkpoint["optimizer"])

    if FLAGS.learning_rate_schedule:
        lr_scheduler = PiecewiseLinear(
            FLAGS.batch_size, [0, 5, FLAGS.epochs], [0, 0.4, 0]
        )
    else:
        lr_scheduler = None

    training_losses = []
    mean_losses = []
    test_losses = []
    accuracies = []

    start = datetime.datetime.now()
    for epoch in range(FLAGS.start_epoch, FLAGS.start_epoch + FLAGS.epochs):
        training_loss, mean_loss = train(
            model,
            device,
            train_loader,
            optimizer,
            epoch,
            lr_scheduler=lr_scheduler,
            writer=writer,
        )
        test_loss, accuracy = test(model, device, test_loader, epoch, writer=writer)

        training_losses += training_loss
        mean_losses.append(mean_loss)
        test_losses.append(test_loss)
        accuracies.append(accuracy)

        if (epoch % FLAGS.model_save_interval == 0) and FLAGS.save_model:
            model_path = f"cifar10-{epoch}.pt"
            save(model_path, model, optimizer)

    stop = datetime.datetime.now()

    np.save("training_losses.npy", np.array(training_losses))
    np.save("mean_losses.npy", np.array(mean_losses))
    np.save("test_losses.npy", np.array(test_losses))
    np.save("accuracies.npy", np.array(accuracies))
    model_path = f"cifar10-final.pt"
    save(model_path, model, optimizer)

    logging.info(f"output saved to {rundir}")
    logging.info(f"{start - stop}")
    writer.close()


if __name__ == "__main__":
    app.run(main)



import torch
import numpy as np

from .logical import logical_and, logical_or, logical_xor, muller_c, posedge_detector


def logical_and_test():
    z = logical_and(torch.as_tensor([1, 0, 0, 1]), torch.as_tensor([1, 0, 0, 0]))
    np.testing.assert_equal(z.numpy(), np.array([1, 0, 0, 0]))


def logical_or_test():
    z = logical_or(torch.as_tensor([1, 0, 0, 1]), torch.as_tensor([1, 0, 0, 0]))
    np.testing.assert_equal(z.numpy(), np.array([1, 0, 0, 1]))


def logical_xor_test():
    z = logical_xor(torch.as_tensor([1, 0, 0, 1]), torch.as_tensor([1, 0, 0, 0]))
    np.testing.assert_equal(z.numpy(), np.array([0, 0, 0, 1]))


def posedge_detector_test():
    z = torch.as_tensor([1, 0, 0, 1])
    z_prev = torch.as_tensor([0, 0, 1, 1])
    z_out = posedge_detector(z, z_prev)
    np.testing.assert_equal(z_out.numpy(), np.array([1, 0, 0, 0]))



import torch


def spiking_vector_quantization_step(v: torch.Tensor, phi: torch.Tensor):
    """Performs the quantization method explained in

    "Deep Spiking Neural Networks", P. O'Connor and Max Welling
    https://arxiv.org/pdf/1602.08323.pdf
    """
    phi = phi + v
    phi_abs = torch.abs(phi)
    phi_abs_sum = torch.sum(phi_abs)
    phi_abs_sorted, indices = torch.sort(phi_abs)
    phi_abs_summed = torch.cumsum(phi_abs_sorted, dim=1)
    z = torch.sign(phi) * torch.sign((phi_abs_sum - phi_abs_summed - 0.5))[indices]
    phi = phi - z
    return z, phi


def constant_spiking_vector_quantization(v: torch.Tensor, T: int):
    phi = torch.zeros_like(v)
    signed_spikes = []

    for _ in range(T):
        z, phi = spiking_vector_quantization_step(v, phi)
        signed_spikes.append(z)

    return torch.stack(signed_spikes)



import importlib
import subprocess
import sys

def main(argv):
    if argv[1] not in ["cifar", "gym", "mnist"]:
        sys.stderr.write("Error: Expected 'cifar', 'gym', or 'mnist'.\n")
        exit(-1)

    task_module = "norse/examples/run_" + argv[1] + ".py"
    subprocess.call([sys.executable, task_module] + argv[2:])


if __name__ == "__main__":
    main(sys.argv)



import torch
import torch.jit

from .lif import LIFState, LIFParameters, lif_step

from .correlation_sensor import (
    CorrelationSensorState,
    CorrelationSensorParameters,
    correlation_sensor_step,
)

from typing import NamedTuple, Tuple


class LIFCorrelationState(NamedTuple):
    lif_state: LIFState
    input_correlation_state: CorrelationSensorState
    recurrent_correlation_state: CorrelationSensorState


class LIFCorrelationParameters(NamedTuple):
    lif_parameters: LIFParameters = LIFParameters()
    input_correlation_parameters: CorrelationSensorParameters = CorrelationSensorParameters()
    recurrent_correlation_parameters: CorrelationSensorParameters = CorrelationSensorParameters()


def lif_correlation_step(
    input: torch.Tensor,
    s: LIFCorrelationState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LIFCorrelationParameters = LIFCorrelationParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFCorrelationState]:
    z_new, s_new = lif_step(
        input, s.lif_state, input_weights, recurrent_weights, p.lif_parameters, dt
    )

    input_correlation_state_new = correlation_sensor_step(
        z_pre=input,
        z_post=z_new,
        s=s.input_correlation_state,
        p=p.input_correlation_parameters,
        dt=dt,
    )

    recurrent_correlation_state_new = correlation_sensor_step(
        z_pre=s.lif_state.z,
        z_post=z_new,
        s=s.recurrent_correlation_state,
        p=p.recurrent_correlation_parameters,
        dt=dt,
    )
    return (
        z_new,
        LIFCorrelationState(
            lif_state=s_new,
            input_correlation_state=input_correlation_state_new,
            recurrent_correlation_state=recurrent_correlation_state_new,
        ),
    )



import setuptools
from setuptools import setup

from os import path

pwd = path.abspath(path.dirname(__file__))

with open(path.join(pwd, "requirements.txt")) as fp:
    install_requires = fp.read()

with open(path.join(pwd, "Readme.md"), encoding="utf-8") as fp:
    readme_text = fp.read()

setup(
    install_requires=install_requires,
    name="norse",
    version="0.0.2",
    description="A library for deep learning with spiking neural networks",
    long_description=readme_text,
    long_description_content_type='text/markdown',
    url='http://github.com/norse/norse',
    author='Christian Pehle',
    author_email='christian.pehle@gmail.com',
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="machine learning spiking neural networks",
)



import torch
import numpy as np
from .heaviside import heaviside


def heaviside_test():
    np.testing.assert_equal(heaviside(torch.ones(100)).numpy(), torch.ones(100).numpy())
    np.testing.assert_equal(
        heaviside(-1.0 * torch.ones(100)).numpy(), torch.zeros(100).numpy()
    )



from typing import NamedTuple, Tuple

import torch
import torch.jit

from .threshhold import threshhold


class LIFParameters(NamedTuple):
    """Parametrization of a LIF neuron

    Parameters:
        tau_syn_inv (torch.Tensor): inverse synaptic time constant (:math:`1/\\tau_\\text{syn}`)
        tau_mem_inv (torch.Tensor): inverse membrane time constant (:math:`1/\\tau_\\text{mem}`)
        v_leak (torch.Tensor): leak potential
        v_th (torch.Tensor): threshhold potential
        v_reset (torch.Tensor): reset potential
        method (str): method to determine the spike threshold (relevant for surrogate gradients)
        alpha (float): hyper parameter to use in surrogate gradient computation
    """

    tau_syn_inv: torch.Tensor = torch.as_tensor(1.0 / 5e-3)
    tau_mem_inv: torch.Tensor = torch.as_tensor(1.0 / 1e-2)
    v_leak: torch.Tensor = torch.as_tensor(0.0)
    v_th: torch.Tensor = torch.as_tensor(1.0)
    v_reset: torch.Tensor = torch.as_tensor(0.0)
    method: str = "super"
    alpha: float = 0.0


class LIFState(NamedTuple):
    """State of a LIF neuron

    Parameters:
        z (torch.Tensor): recurrent spikes
        v (torch.Tensor): membrane potential
        i (torch.Tensor): synaptic input current
    """

    z: torch.Tensor
    v: torch.Tensor
    i: torch.Tensor


class LIFFeedForwardState(NamedTuple):
    """State of a feed forward LIF neuron

    Parameters:
        v (torch.Tensor): membrane potential
        i (torch.Tensor): synaptic input current
    """

    v: torch.Tensor
    i: torch.Tensor


def lif_step(
    input: torch.Tensor,
    s: LIFState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFState]:
    r"""Computes a single euler-integration step of a LIF neuron-model. More
    specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + w_{\text{input}} z_{\text{in}} \\
            i &= i + w_{\text{rec}} z_{\text{rec}}
        \end{align*}

    where :math:`z_{\text{rec}}` and :math:`z_{\text{in}}` are the recurrent and input
    spikes respectively.

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFState): current state of the LIF neuron
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LIFParameters): parameters of a leaky integrate and fire neuron
        dt (float): Integration timestep to use
    """
    # compute voltage updates
    dv = dt * p.tau_mem_inv * ((p.v_leak - s.v) + s.i)
    v_decayed = s.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * s.i
    i_decayed = s.i + di

    # compute new spikes
    z_new = threshhold(v_decayed - p.v_th, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = (
        i_decayed
        + torch.nn.functional.linear(input, input_weights)
        + torch.nn.functional.linear(s.z, recurrent_weights)
    )

    return z_new, LIFState(z_new, v_new, i_new)


def lif_feed_forward_step(
    input: torch.Tensor,
    s: LIFFeedForwardState = LIFFeedForwardState(0, 0),
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFFeedForwardState]:
    r"""Computes a single euler-integration step for a lif neuron-model.
    It takes as input the input current as generated by an arbitrary torch module
    or function. More specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + i_{\text{in}}
        \end{align*}

    where :math:`i_{\text{in}}` is meant to be the result of applying an arbitrary
    pytorch module (such as a convolution) to input spikes.

    Parameters:
        input (torch.Tensor): the input spikes at the current time step
        s (LIFFeedForwardState): current state of the LIF neuron
        p (LIFParameters): parameters of a leaky integrate and fire neuron
        dt (float): Integration timestep to use
    """
    # compute voltage updates
    dv = dt * p.tau_mem_inv * ((p.v_leak - s.v) + s.i)
    v_decayed = s.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * s.i
    i_decayed = s.i + di

    # compute new spikes
    z_new = threshhold(v_decayed - p.v_th, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = i_decayed + input

    return z_new, LIFFeedForwardState(v_new, i_new)


def lif_current_encoder(
    input_current: torch.Tensor,
    v: torch.Tensor,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""Computes a single euler-integration step of a leaky integrator. More
    specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    Parameters:
        input (torch.Tensor): the input current at the current time step
        s (LIFState): current state of the LIF neuron
        p (LIFParameters): parameters of a leaky integrate and fire neuron
        dt (float): Integration timestep to use
    """
    dv = dt * p.tau_mem_inv * ((p.v_leak - v) + input_current)
    v = v + dv
    z = threshhold(v - p.v_th, p.method, p.alpha)

    v = v - z * (v - p.v_reset)
    return z, v



import torch
import norse.torch.module.lsnn as lsnn

from nose.tools import raises


def test_lsnn_cell():
    cell = lsnn.LSNNCell(2, 10)
    state = cell.initial_state(5, "cpu")
    data = torch.zeros(5, 2)
    z, state = cell(data, state)
    assert torch.equal(z, torch.zeros(5, 10))


@raises(TypeError)
def test_lsnn_cell_param_fail():
    cell = lsnn.LSNNCell()


@raises(TypeError)
def test_lsnn_state_fail():
    cell = lsnn.LSNNCell(2, 10)
    cell.initial_state()


@raises(RuntimeError)
def test_lsnn_forward_shape_fail():
    cell = lsnn.LSNNCell(2, 10)
    state = cell.initial_state(5, "cpu")
    data = torch.zeros(10)
    cell.forward(data, state)


def test_lsnn_layer():
    layer = lsnn.LSNNLayer(lsnn.LSNNCell, 2, 10)
    state = layer.initial_state(5, "cpu")
    data = torch.zeros(2, 5, 2)
    z, state = layer.forward(data, state)
    assert torch.equal(z, torch.zeros(2, 5, 10))
